#!/usr/bin/env python

import re
import sys
import numpy as np
import math
from scipy import optimize
import random
from scipy.stats import pearsonr


#########################
# Specify path, file names and values of experimentally determined variables
#########################

# path, where the input files are stored
path = str(sys.argv[1])

# file, which contains read counts associated with different site types in various binding reactions
reads_file_name = str(sys.argv[2])

# total concentration of RNA used in the binding reaction (nM)
total_RNA_concentration = float(sys.argv[3])

# RISC concentration
a = float(sys.argv[4])

leave_one_out_do = int(sys.argv[5])

#########################
# Get read counts of each site type in binding reactions and dilution_factors (experimental data)
# Get total concentration of each site type in the library (experimental data)
#########################
# reads_it represents the read counts associated with site type i at a time point t.
# They are stored in a tab file, which contains m rows (site types) and n+2 columns
# (n time points + first column describing the site types + second column counting reads in the Input)

read_count_data=open(path+reads_file_name,'r')

line = read_count_data.readline()
time_points = []
elements = re.split(r'\t',re.split(r'\n',line)[0])

for i in range(len(elements)):
    if i != 0 and i != 1:
        time_points.append(float(elements[i]))

read_counts_all = {}
input_concentrations={}
line = read_count_data.readline()

while line:
    site_type=re.split(r'\t',line)[0]
    site_read_counts=[]
    for i in range(len(time_points)):
        site_read_counts.append(float(re.split(r'\t',re.split(r'\n',line)[0])[i+2]))
    read_counts_all.update({site_type: site_read_counts})
    input_concentrations.update({re.split(r'\t',line)[0]: float(re.split(r'\t',line)[1])})
    line=read_count_data.readline()
read_count_data.close()

total=sum(input_concentrations[key] for key in input_concentrations.keys())
for key in input_concentrations.keys():
    input_concentrations.update({key: input_concentrations[key] / float(total) * float(total_RNA_concentration)})

Input = sum(input_concentrations[i] for i in input_concentrations.keys())

#########################
# Definition of sit, which is the concentration of site type i bound at time t (mathematical model)
#########################

def sit(t,koni,koffi,afreet,inputi):
    # inputi = concentration of site type i in the input library.
    # kon = dictionary of kon_i values for all site types; has m elements.
    # koff = dictionary of koff_i values for all site types; has m elements.
    # i is the site_type and t is the time point.
    # afree is concentration of free RISC.
        
    value = koni * inputi * afreet / (koni * afreet + koffi) - koni * inputi * a / (koni * a + koffi) * np.exp(-(koni * afreet + koffi)*t)
    return value

#########################
# Definition of xit, which is the concentration of site type i recovered in binding reaction j (mathematical model)
#########################

def xit(t,inputi,Input,background,sit,St):
    # inputi = concentration of site type i in the input library.
    # i is the site_type and t is the time point that I am calculating xit for
    # background = total concentration of nonspecifically recovered RNA, assumed to be constant across binding reactions
    # sit is the concentration of site type i bound at time t (mathematical model)
    
    value = sit * (1 - background / (Input - St)) + background * inputi / (Input - St)
    return value

#########################
# Calculation of afree at time t 
#########################

# a_unbound is computed at each iteration cycle of the optimization process.

def free(t,kon,koff,a,input_concentrations):
    
    def equation_afree(x):
        return a - x - sum(kon[z]*input_concentrations[z]*x/(kon[z]*x+koff[z])-kon[z]*input_concentrations[z]*a/(kon[z]*a+koff[z])*np.exp(-(kon[z]*x+koff[z])*t) for z in kon.keys())
    
    res = optimize.minimize_scalar(equation_afree, bounds=(0, a), method='bounded')
    #res = optimize.minimize(equation_afree, 0.15, bounds=(0, a))
    return res.x

#########################
# Important note on theta and initialization
#########################

# For syntax purposes, theta cannot be a dictionary, but can be a list.
# Therefore, theta will be arranged in a certain order and corresponding indices will be stored in index_list.
m = len(input_concentrations.keys())
index_list = [None] * (2*m)
temp = input_concentrations.keys()
temp.remove('nosite')
temp.append('nosite')
for i in range(len(temp)):
    index_list[i] = temp[i]+'.kon'
    index_list[i+m] = temp[i]+'.koff'
index_list.append('background')

def initialization(read_counts,teq):
    # Compute initial guess of theta
    # at the same time, create a list of bounds for optimization
    theta_0_array=np.array([])
    m = len(read_counts.keys())
    for i in range(m-1):
        site = re.split(r'.k',index_list[i])[0]
        R0 = sum(read_counts[z][0] for z in read_counts.keys()) 
        freq_RISC_site0 = float(read_counts[site][0]) / float(R0)
        enrichment0 = np.mean(freq_RISC_site0) / (input_concentrations[site] / total_RNA_concentration)
        R3 = sum(read_counts[z][3] for z in read_counts.keys()) 
        freq_RISC_site3 = float(read_counts[site][3]) / float(R3)
        enrichment3 = np.mean(freq_RISC_site3) / (input_concentrations[site] / total_RNA_concentration)
        kon=(enrichment3-enrichment0)/a/(time_points[3]-time_points[0])
        theta_0_array = np.append(theta_0_array, np.log(abs(kon)+np.random.normal(loc=0,scale=abs(kon),size=None)))
    theta_0_array = np.append(theta_0_array, np.log(0.1+np.random.normal(loc=0,scale=0.001,size=None)))
    for i in range(m-1):
        site = re.split(r'.k',index_list[i])[0]
        R = sum(read_counts[z][teq] for z in read_counts.keys())  # total number of reads at equilibrium (last time point)
        freq_RISC_site = float(read_counts[site][teq]) / float(R)
        enrichment = np.mean(freq_RISC_site) / (input_concentrations[site] / total_RNA_concentration)
        theta_0_array = np.append(theta_0_array, np.log(1/enrichment+np.random.normal(loc=0,scale=1/enrichment/100,size=None))) # Compute koff as 1/enrichment*a
    theta_0_array = np.append(theta_0_array, np.log(10+np.random.normal(loc=0,scale=1,size=None))) #koff nosite at 100 s-1
    theta_0_array = np.append(theta_0_array, np.log(0.5+np.random.normal(loc=0,scale=0.1,size=None))) #background is 0.5 nM + a random value drawn from a normal distribution with mean 0 and standard deviation 0.1
    return theta_0_array

#########################
# Defining mathematical model of concentrations recovered for all m site types in all n binding reactions 
#########################

# The output is a dictionary, which contains predicted concentrations for all m site types in all n binding reactions
# The structure of the dictionary is similar to read_counts:
# each site type is a key and the value is a list of length n of values for different binding reactions.
# The input is a list of parameters theta, which are the goal of optimization procedure:
# KD1 = exp(theta1), KD2 = exp(theta2), ..., KD_m = exp(theta_m)
# a = exp(theta_m+1), where a is stock concentration of RISC
# background = exp(theta_m+2).

def values_from_theta(theta,index_list):
    kon = {}
    koff = {}
    for z in range(len(theta)-1):
        if re.split(r'.k',index_list[z])[1] == 'on':
            kon.update({re.split(r'.k',index_list[z])[0]: np.exp(theta[z])})
        elif re.split(r'.k',index_list[z])[1] == 'off':
            koff.update({re.split(r'.k',index_list[z])[0]: np.exp(theta[z])})
    return kon, koff

def model(theta):
    # Step 1: compute afree. 
    kon, koff = values_from_theta(theta,index_list)
    afree = []
    for t in time_points:
        afree.append(free(t,kon,koff,a,input_concentrations))
    
    # Step 2: calculate sit for all the site types at all time points. This will yield 
    # a dictionary of m keys and n-long lists of values.
    sit_dict = {}
    for i in input_concentrations.keys():
        temp=[]
        for t in range(len(time_points)):
            temp.append(sit(t,kon[i],koff[i],afree[t],input_concentrations[i]))
        sit_dict.update({i: temp})
    
    S=[]
    for t in range(len(time_points)):
        S.append(sum(sit_dict[i][t] for i in sit_dict.keys()))
    
    model_prediction = {}
    background = np.exp(theta[len(theta) - 1])
    for i in input_concentrations.keys():
        temp=[]
        for t in range(len(time_points)):
            temp.append(xit(t,input_concentrations[i],Input,background,sit_dict[i][t],S[t]))
        model_prediction.update({i: temp})
    
    return sit_dict, model_prediction

#########################
# Definition of fcost 
#########################

def fcost(theta_array, scalar=0.00000001):
    theta=theta_array.tolist()
    sit_dict, model_prediction = model(theta)
    R=[]  # Rj is total number of reads of all m sites at a time point j (experimental data)
    X=[]  # Xj is total concentration of all m sites at a time point j (mathematical model)
    S=[]  # Sj is sum (over all site types) of reads_ij*ln(xij)
    for t in range(len(points)):
        R.append(sum(read_counts[key][t] for key in read_counts.keys()))
        X.append(sum(model_prediction[i][t] for i in model_prediction.keys()))
        S.append(sum(read_counts[i][t]*np.log(model_prediction[i][t]) for i in read_counts.keys()))
    S_temp=sum(R[t]*np.log(X[t])-S[t] for t in range(len(points)))
    return scalar * sum(R[t]*np.log(X[t])-S[t] for t in range(len(points)))

#########################
# Definition of gradient
#########################

def gradient(theta_array):
    theta=theta_array.tolist()
    sit_dict, model_prediction = model(theta)
    kon, koff = values_from_theta(theta,index_list)
    m = len(kon.keys())-1
    background = np.exp(theta[len(theta) - 1])
    der=np.array([])
    for k in range(len(theta)):
        # Start the calculation of d(fcost)/d(theta_k)
        if k == len(theta)-1:
            delta_k_2m1 = 1
        else:
            delta_k_2m1 = 0
        der_k = 0
        for t in range(len(points)):
            afree = free(t,kon,koff,a,input_concentrations)
            Rt = sum(read_counts[key][t] for key in read_counts.keys())
            Xt = sum(model_prediction[i][t] for i in model_prediction.keys())
            St = sum(sit_dict[i][t] for i in sit_dict.keys())
            for z in range(m):
                i = re.split(r'.k',index_list[z])[0]
                A = background * (input_concentrations[i] - sit_dict[i][t]) / (total_RNA_concentration - St) * delta_k_2m1
                B = background * (input_concentrations[i] - sit_dict[i][t]) / (total_RNA_concentration - St)**2
                C = 1 - background / (total_RNA_concentration - St)
                D = 0
                if k == z:
                    D1 = input_concentrations[i] * afree * koff[i] / (kon[i] * afree + koff[i])**2
                    D2 = input_concentrations[i] * a / (kon[i] * a + koff[i]) * np.exp(-t*(kon[i]*afree+koff[i]))
                    D3 = koff[i] / (kon[i]*a+koff[i]) - afree*t*kon[i]
                    D += D1 - D2 * D3
                elif k == m+z:
                    D1 = kon[i] * input_concentrations[i] * afree / (kon[i] * afree + koff[i])**2
                    D2 = kon[i] * input_concentrations[i] * a / (kon[i] * a + koff[i]) * np.exp(-t*(kon[i]*afree+koff[i]))
                    D3 = t - 1 / (kon[i] * a + koff[i])
                    D += -D1 - D2 * D3
                E1 = kon[i] * input_concentrations[i] * koff[i] / (kon[i] * afree + koff[i])**2
                E2 = kon[i]**2 * input_concentrations[i] * a * t / (kon[i] * a + koff[i]) * np.exp(-t*(kon[i]*afree+koff[i]))
                E = E1 + E2
                F = 0
                for y in sit_dict.keys():
                    E1 = kon[y] * input_concentrations[y] * koff[y] / (kon[y] * afree + koff[y])**2
                    E2 = kon[y]**2 * input_concentrations[y] * a * t / (kon[y] * a + koff[y]) * np.exp(-t*(kon[y]*afree+koff[y]))
                    F += E1 + E2
                if k <= m:
                    D1 = input_concentrations[i] * afree * koff[i] / (kon[i] * afree + koff[i])**2
                    D2 = input_concentrations[i] * a / (kon[i] * a + koff[i]) * np.exp(-t*(kon[i]*afree+koff[i]))
                    D3 = koff[i] / (kon[i]*a+koff[i]) - afree*t*kon[i]
                    G = (D1 - D2 * D3)/(1+F)
                    D += -E*(D1 - D2 * D3)/(1+F)
                elif k > m and k <= 2*m:
                    D1 = kon[i] * input_concentrations[i] * afree / (kon[i] * afree * koff[i])**2
                    D2 = kon[i] * input_concentrations[i] * a / (kon[i] * a + koff[i]) * np.exp(-t*(kon[i]*afree+koff[i]))
                    D3 = t - 1 / (kon[i] * a + koff[i])
                    G = (-D1 - D2 * D3)/(1+F)
                    D += -E*(-D1 - D2 * D3)/(1+F)
                der_k += (Rt / Xt - read_counts[i][t] / model_prediction[i][t]) * (A + C * D + B * G)
        der = np.append(der, der_k)
        
    return der

#########################
# Prediction of motif counts given parameters values 
#########################

def predict_concentrations(theta_array,input_concentrations,time_points):
    theta=theta_array.tolist()
    kon, koff = values_from_theta(theta,index_list)
    background = np.exp(theta[len(theta) - 1])
    predicted_concentrations = {}
    for t in range(len(time_points)):
        afree = free(t,kon,koff,a,input_concentrations)
        sit_dict = {}
        for i in kon.keys():
            sit = kon[i] * input_concentrations[i] * afree / (kon[i] * afree + koff[i]) * (1 - np.exp(-time_points[t]*(kon[i]*afree+koff[i])))
            sit_dict.update({i: sit})
        St = sum(sit_dict[z] for z in sit_dict.keys())
        Input = sum(input_concentrations[z] for z in input_concentrations.keys())
        Pr = []
        Rj = sum(read_counts[z][t] for z in read_counts.keys())  # total number of reads in a time point t (experimental data)
        for i in kon.keys():
            xit = sit_dict[i]*(1-background/(Input - St))+background*input_concentrations[i]/(Input - St)
            Pr.append(float(xit))
        Prj = sum(z for z in Pr)
        temp = []
        for m in Pr:
            temp.append(m/Prj*Rj)
        predicted_concentrations.update({time_points[t]: temp})

    predicted_concentrations_transposed = {}
    c = 0
    for i in kon.keys():
        temp = []
        for t in range(len(time_points)):
            temp.append(predicted_concentrations[time_points[t]][c])
        predicted_concentrations_transposed.update({i: temp})
        c += 1
    return predicted_concentrations_transposed

#########################
# Leave-one-out command 
#########################

def leave_one_out(read_count_dict,index_to_remove):
    ans = {}
    for site_type in read_count_dict.keys():
        value = []
        for v in range(len(read_count_dict[site_type])):
            if v != index_to_remove:
                value.append(read_count_dict[site_type][v])
        ans.update({site_type: value})
    return ans

#########################
# Optimization (repeated noptimize times) command 
#########################

noptimize = 0
list_bounds = []
for i in range(m):
    list_bounds.append((np.log(0.0001),np.log(1000))) # bondaries for kon, including no-site
for i in range(m-1):
    list_bounds.append((np.log(0.0005),np.log(20))) # bondaries for koff
list_bounds.append((np.log(0.0005),np.log(200))) # bondaries for koff of no-site
list_bounds.append((np.log(0.01),np.log(10))) # # bondaries for background

output = open(reads_file_name+'_kon_all','w')
for i in range(len(index_list)):
    output.write(str(index_list[i])+'\t')
output.write('\n')

while noptimize == 0:
    read_counts = read_counts_all.copy()
    points = list(time_points)
    theta_0_array = initialization(read_counts, len(points)-1)
    print 'Starting optimization using all time points'+'\n'
    res = optimize.minimize(fcost, theta_0_array, bounds = list_bounds, method='L-BFGS-B', options={'disp': False})
    if res.success == False:
        print "Partially initialized values do not allow calculating fcost."
    if res.success == True:
        prediction = predict_concentrations(res.x,input_concentrations,time_points)
        flag = 1
        for j in range(len(time_points)):
            predicted_counts = []
            sequenced_counts = []
            for i in prediction.keys():
                predicted_counts.append(prediction[i][j])
                sequenced_counts.append(read_counts[i][j])
            corr, _ = pearsonr(predicted_counts, sequenced_counts)
            if corr < 0.90:
                flag = 0
                print 'Pearson r between the model and the data is below 0.90.'+'\n'
        if flag == 1:            
            noptimize=1

for i in range(len(index_list)):
    output.write(str(np.exp(res.x[i]))+'\t')
output.write('\n')
output.close()
print 'File '+reads_file_name+'_kon_all'+' contains MLE results.'


if leave_one_out_do == 1:
    for j in range(len(time_points)):
        output = open(reads_file_name+'_kon_wo_'+str(time_points[j]),'w')
        for i in range(len(index_list)):
            output.write(str(index_list[i])+'\t')
        output.write('\n')
        read_counts = leave_one_out(read_counts_all,j)
        points = list(time_points)
        del points[j]
        flag = 0
        while flag == 0:
            theta_0_array = initialization(read_counts, len(points)-1)
            print str(time_points[j])+' removed. Starting optimization'+'\n'
            res = optimize.minimize(fcost, theta_0_array, bounds = list_bounds, method='L-BFGS-B', options={'disp': False})
            if res.success == False:
                print "Partially initialized values do not allow calculating fcost."
            if res.success == True:
                prediction = predict_concentrations(res.x,input_concentrations,points)
                flag = 1
                for z in range(len(points)):
                    predicted_counts = []
                    sequenced_counts = []
                    for i in prediction.keys():
                        predicted_counts.append(prediction[i][z])
                        sequenced_counts.append(read_counts[i][z])
                    corr, _ = pearsonr(predicted_counts, sequenced_counts)
                    if corr < 0.90:
                        flag = 0
                        print 'Pearson r between the model and the data is below 0.90.'+'\n'
        for i in range(len(index_list)):
            output.write(str(np.exp(res.x[i]))+'\t')
        output.write('\n')
        output.close()
        print 'File '+reads_file_name+'_kon_wo_'+str(time_points[j])+' contains MLE results.'


